@@ -4,7 +4,7 @@ use anyhow::{bail, Result};
 use std::{collections::HashSet, ffi::OsStr, path::Path, process::Command};
 
 use crate::{
-    backend::{Backend, Branch, Config, HostBranch, Remote, Snapshot},
+    backend::{Backend, Branch, Config, HostBranch, PruneFrom, Remote, RemoteHostBranch, Snapshot},
     git_ref::GitRef,
     progress::{output_stdout, Progress, Run},
 };
@@ -38,7 +38,7 @@ fn git_command<S: AsRef<OsStr>>(name: S) -> Command {
 /// both built-in and third party.
 mod namespace {
     use crate::{
-        backend::{Branch, Config, HostBranch},
+        backend::{Branch, Config, HostBranch, RemoteHostBranch},
         git_ref::GitRef,
     };
 
@@ -51,6 +51,15 @@ mod namespace {
         format!("{}.{}", PREFIX, key)
     }
 
+    /// The refspec to list remote nomad managed refs.
+    pub fn list_refspec(config: &Config) -> String {
+        format!(
+            "refs/{prefix}/{user}/*",
+            prefix = PREFIX,
+            user = config.user
+        )
+    }
+
     /// The refspec to fetch remote nomad managed refs as local refs.
     ///
     /// `refs/nomad/rraval/apollo/master` becomes `refs/nomad/apollo/master`.
@@ -58,9 +67,9 @@ mod namespace {
     /// `refs/nomad/rraval/boreas/feature` becomes `refs/nomad/boreas/feature`.
     pub fn fetch_refspec(config: &Config) -> String {
         format!(
-            "+refs/{prefix}/{user}/*:refs/{prefix}/*",
+            "+{remote_pattern}:refs/{prefix}/*",
+            remote_pattern = list_refspec(config),
             prefix = PREFIX,
-            user = config.user
         )
     }
 
@@ -94,7 +103,7 @@ mod namespace {
         format!("refs/{}/{}/{}/{}", PREFIX, config.user, config.host, branch)
     }
 
-    pub fn as_host_branch(git_ref: GitRef) -> Result<HostBranch<GitRef>, GitRef> {
+    pub fn host_branch_from_local_ref(git_ref: GitRef) -> Result<HostBranch<GitRef>, GitRef> {
         let parts = git_ref.name.split('/').collect::<Vec<_>>();
         match parts.as_slice() {
             ["refs", prefix, host, branch_name] => {
@@ -112,19 +121,39 @@ mod namespace {
         }
     }
 
+    pub fn host_branch_from_remote_ref(git_ref: GitRef) -> Result<RemoteHostBranch, GitRef> {
+        let parts = git_ref.name.split('/').collect::<Vec<_>>();
+        match parts.as_slice() {
+            ["refs", prefix, user, host, branch_name] => {
+                if prefix != &PREFIX {
+                    return Err(git_ref);
+                }
+
+                Ok(RemoteHostBranch {
+                    user: user.to_string(),
+                    host: host.to_string(),
+                    branch: Branch::str(branch_name),
+                })
+            }
+            _ => Err(git_ref),
+        }
+    }
+
     #[cfg(test)]
     mod tests {
         use crate::{
-            backend::{Branch, Config, HostBranch},
+            backend::{Branch, Config, HostBranch, RemoteHostBranch},
             git_ref::GitRef,
         };
 
-        use super::{as_host_branch, local_ref};
+        use super::{
+            host_branch_from_local_ref, host_branch_from_remote_ref, local_ref, remote_ref,
+        };
 
-        /// [`as_host_branch`] should be able to parse ref names produced by [`local_ref`] (they
+        /// [`host_branch_from_local_ref`] should be able to parse ref names produced by [`local_ref`] (they
         /// are duals).
         #[test]
-        fn test_create_and_parse() {
+        fn test_create_and_parse_local_ref() {
             let git_ref = GitRef {
                 commit_id: "some_commit_id".to_string(),
                 name: local_ref(
@@ -137,7 +166,7 @@ mod namespace {
             };
 
             assert_eq!(
-                as_host_branch(git_ref.clone()),
+                host_branch_from_local_ref(git_ref.clone()),
                 Ok(HostBranch {
                     host: "host0".to_string(),
                     branch: Branch::str("branch0"),
@@ -145,6 +174,31 @@ mod namespace {
                 })
             );
         }
+
+        /// [`host_branch_from_remote_ref`] should be able to parse ref names produced by
+        /// [`remote_ref`] (they are duals).
+        #[test]
+        fn test_create_and_parse_remote_ref() {
+            let git_ref = GitRef {
+                commit_id: "some_commit_id".to_string(),
+                name: remote_ref(
+                    &Config {
+                        user: "user0".to_string(),
+                        host: "host0".to_string(),
+                    },
+                    "branch0",
+                ),
+            };
+
+            assert_eq!(
+                host_branch_from_remote_ref(git_ref),
+                Ok(RemoteHostBranch {
+                    user: "user0".to_string(),
+                    host: "host0".to_string(),
+                    branch: Branch::str("branch0"),
+                })
+            );
+        }
     }
 }
 
@@ -242,7 +296,7 @@ impl<'progress, 'name> GitBinary<'progress, 'name> {
     fn fetch_refspecs<Description, RefSpec>(
         &self,
         description: Description,
-        remote: &str,
+        remote: &Remote,
         refspecs: &[RefSpec],
     ) -> Result<()>
     where
@@ -253,7 +307,7 @@ impl<'progress, 'name> GitBinary<'progress, 'name> {
         self.progress.run(
             Run::Notable,
             description,
-            self.command().args(&["fetch", remote]).args(refspecs),
+            self.command().args(&["fetch", &remote.0]).args(refspecs),
         )?;
         Ok(())
     }
@@ -267,7 +321,7 @@ impl<'progress, 'name> GitBinary<'progress, 'name> {
     fn push_refspecs<Description, RefSpec>(
         &self,
         description: Description,
-        remote: &str,
+        remote: &Remote,
         refspecs: &[RefSpec],
     ) -> Result<()>
     where
@@ -278,7 +332,7 @@ impl<'progress, 'name> GitBinary<'progress, 'name> {
         self.progress.run(
             Run::Notable,
             description,
-            self.command().args(&["push", remote]).args(refspecs),
+            self.command().args(&["push", &remote.0]).args(refspecs),
         )?;
         Ok(())
     }
@@ -322,6 +376,39 @@ impl<'progress, 'name> GitBinary<'progress, 'name> {
             .collect()
     }
 
+    /// Wraps `git ls-remote` to query a remote for all refs that match the given `refspecs`.
+    ///
+    /// # Panics
+    ///
+    /// If `refspecs` is empty, which means git will list all refs, which is never what we want.
+    fn list_remote_refs<Description, RefSpec>(
+        &self,
+        description: Description,
+        remote: &Remote,
+        refspecs: &[RefSpec],
+    ) -> Result<Vec<GitRef>>
+    where
+        Description: AsRef<str>,
+        RefSpec: AsRef<OsStr>,
+    {
+        assert!(!refspecs.is_empty());
+        let output = self
+            .progress
+            .run(
+                Run::Notable,
+                description,
+                self.command()
+                    .arg("ls-remote")
+                    .arg(&remote.0)
+                    .args(refspecs),
+            )
+            .and_then(output_stdout)?;
+        output
+            .lines()
+            .map(|line| GitRef::parse_ls_remote_line(line).map_err(Into::into))
+            .collect()
+    }
+
     /// Delete a ref from the repository.
     ///
     /// Note that deleting refs on a remote is done via [`GitBinary::push_refspecs`].
@@ -375,7 +462,7 @@ impl<'progress, 'name> Backend for GitBinary<'progress, 'name> {
                 local_branches.insert(Branch::str(name));
             }
 
-            if let Ok(host_branch) = namespace::as_host_branch(r) {
+            if let Ok(host_branch) = namespace::host_branch_from_local_ref(r) {
                 host_branches.push(host_branch);
             }
         }
@@ -386,43 +473,64 @@ impl<'progress, 'name> Backend for GitBinary<'progress, 'name> {
         })
     }
 
-    fn fetch(&self, config: &Config, remote: &Remote) -> Result<()> {
+    fn fetch(&self, config: &Config, remote: &Remote) -> Result<HashSet<RemoteHostBranch>> {
         self.fetch_refspecs(
             format!("Fetching branches from {}", remote.0),
-            &remote.0,
+            remote,
             &[&namespace::fetch_refspec(config)],
-        )
+        )?;
+
+        // In an ideal world, we would be able to get the list of refs fetched directly from `git`.
+        //
+        // However, `git fetch` is a porcelain command and we don't want to get into parsing its
+        // output, so do an entirely separate network fetch with the plumbing `git ls-remote` which
+        // we can parse instead.
+        let remote_refs = self.list_remote_refs(
+            format!("Listing branches at {}", remote.0),
+            remote,
+            &[&namespace::list_refspec(config)],
+        )?;
+
+        Ok(remote_refs
+            .into_iter()
+            .filter_map(|ref_| namespace::host_branch_from_remote_ref(ref_).ok())
+            .collect())
     }
 
     fn push(&self, config: &Config, remote: &Remote) -> Result<()> {
         self.push_refspecs(
             format!("Pushing local branches to {}", remote.0),
-            &remote.0,
+            remote,
             &[&namespace::push_refspec(config)],
         )
     }
 
     fn prune<'b, Prune>(&self, config: &Config, remote: &Remote, prune: Prune) -> Result<()>
     where
-        Prune: Iterator<Item = &'b HostBranch<GitRef>>,
+        Prune: Iterator<Item = &'b PruneFrom<GitRef>>,
     {
         let mut refspecs = Vec::<String>::new();
         let mut refs = Vec::<GitRef>::new();
 
-        for host_branch in prune {
-            refspecs.push(format!(
-                ":{}",
-                namespace::remote_ref(config, &host_branch.branch.0)
-            ));
+        for prune_from in prune {
+            if let PruneFrom::LocalAndRemote(hb) = prune_from {
+                refspecs.push(format!(":{}", namespace::remote_ref(config, &hb.branch.0)));
+            }
 
-            refs.push(host_branch.ref_.clone());
+            refs.push(
+                match prune_from {
+                    PruneFrom::LocalOnly(hb) | PruneFrom::LocalAndRemote(hb) => hb,
+                }
+                .ref_
+                .clone(),
+            );
         }
 
         // Delete from the remote first
         if !refspecs.is_empty() {
             self.push_refspecs(
                 format!("Pruning branches at {}", remote.0),
-                &remote.0,
+                remote,
                 &refspecs,
             )?;
         }
@@ -588,14 +696,14 @@ mod test_backend {
     use std::{
         collections::HashSet,
         fs::{create_dir, write},
-        iter,
+        iter::FromIterator,
         path::PathBuf,
     };
 
     use tempfile::{tempdir, TempDir};
 
     use crate::{
-        backend::{Backend, Branch, Config, HostBranch, Remote},
+        backend::{Backend, Branch, Config, HostBranch, PruneFrom, Remote, RemoteHostBranch},
         git_binary::namespace,
         progress::{Progress, Run, Verbosity},
     };
@@ -604,7 +712,7 @@ mod test_backend {
 
     const GIT: &str = "git";
     const ORIGIN: &str = "origin";
-    const BRANCH: &str = "branch0";
+    const INITIAL_BRANCH: &str = "master";
     const USER: &str = "user0";
 
     const PROGRESS: Progress = Progress::Verbose(Verbosity::CommandAndOutput);
@@ -634,7 +742,7 @@ mod test_backend {
                 };
 
                 create_dir(remote_dir).unwrap();
-                git(&["init", "--initial-branch", BRANCH]);
+                git(&["init", "--initial-branch", INITIAL_BRANCH]);
 
                 let file0 = remote_dir.join("file0");
                 write(file0, "line0\nline1\n").unwrap();
@@ -703,24 +811,27 @@ mod test_backend {
             self.git.push(&self.config, &self.remote()).unwrap();
         }
 
-        fn fetch(&self) {
+        fn fetch(&self) -> HashSet<RemoteHostBranch> {
             self.git.fetch(&self.config, &self.remote()).unwrap()
         }
 
-        fn prune(&self) {
-            let ref_name = namespace::local_ref(&self.config, BRANCH);
-            let git_ref = self.git.get_ref("", ref_name).unwrap();
+        fn prune_local_and_remote<'b, B: IntoIterator<Item = &'b str>>(&self, branch_names: B) {
+            let prune_from: Vec<_> = branch_names
+                .into_iter()
+                .map(|name| {
+                    let ref_name = namespace::local_ref(&self.config, name);
+                    let ref_ = self.git.get_ref("", ref_name).unwrap();
+
+                    PruneFrom::LocalAndRemote(HostBranch {
+                        host: self.config.host.clone(),
+                        branch: Branch::str(name),
+                        ref_,
+                    })
+                })
+                .collect();
 
             self.git
-                .prune(
-                    &self.config,
-                    &self.remote(),
-                    iter::once(&HostBranch {
-                        host: self.config.host.clone(),
-                        branch: Branch::str(BRANCH),
-                        ref_: git_ref,
-                    }),
-                )
+                .prune(&self.config, &self.remote(), prune_from.iter())
                 .unwrap();
         }
     }
@@ -749,8 +860,8 @@ mod test_backend {
 
         assert_eq!(ref_names(&refs), {
             let mut set: HashSet<String> = HashSet::new();
-            set.insert(format!("refs/heads/{}", BRANCH));
-            set.insert(namespace::remote_ref(&host0.config, BRANCH));
+            set.insert(format!("refs/heads/{}", INITIAL_BRANCH));
+            set.insert(namespace::remote_ref(&host0.config, INITIAL_BRANCH));
             set
         });
 
@@ -774,10 +885,10 @@ mod test_backend {
             let mut set: HashSet<String> = HashSet::new();
 
             // the local branch
-            set.insert(format!("refs/heads/{}", BRANCH));
+            set.insert(format!("refs/heads/{}", INITIAL_BRANCH));
 
             // the remote branch and what's checked out
-            set.insert(format!("refs/remotes/{}/{}", ORIGIN, BRANCH));
+            set.insert(format!("refs/remotes/{}/{}", ORIGIN, INITIAL_BRANCH));
             set.insert(format!("refs/remotes/{}/HEAD", ORIGIN));
 
             set
@@ -797,7 +908,15 @@ mod test_backend {
                 .branches_for_host(&host0.config),
             vec![]
         );
-        host1.fetch();
+        let remote_host_branches = host1.fetch();
+        assert_eq!(
+            remote_host_branches,
+            HashSet::from_iter([RemoteHostBranch {
+                user: USER.to_string(),
+                host: "host0".to_string(),
+                branch: Branch::str(INITIAL_BRANCH),
+            }]),
+        );
 
         // After fetch, we should have the additional ref
         {
@@ -805,7 +924,7 @@ mod test_backend {
             assert_eq!(ref_names(&refs), {
                 let mut set = pre_fetch_refs();
                 // the additional ref from the host0 push
-                set.insert(namespace::local_ref(&host0.config, BRANCH));
+                set.insert(namespace::local_ref(&host0.config, INITIAL_BRANCH));
                 set
             });
             assert_eq!(
@@ -814,7 +933,7 @@ mod test_backend {
                     .snapshot()
                     .unwrap()
                     .branches_for_host(&host0.config),
-                vec![Branch::str(BRANCH)]
+                vec![Branch::str(INITIAL_BRANCH)]
             );
         }
     }
@@ -858,7 +977,7 @@ mod test_backend {
         let empty_set = HashSet::new();
         let branch_set = {
             let mut set = HashSet::new();
-            set.insert(BRANCH.to_string());
+            set.insert(INITIAL_BRANCH.to_string());
             set
         };
 
@@ -877,7 +996,7 @@ mod test_backend {
         assert_eq!(local_nomad_refs(), branch_set);
 
         // Pruning removes the ref remotely and locally
-        local.prune();
+        local.prune_local_and_remote([INITIAL_BRANCH]);
         assert_eq!(remote_nomad_refs(), empty_set);
         assert_eq!(local_nomad_refs(), empty_set);
     }
